
/* cwds.c - wrapper for CellWriter

Copyright (C) 2009 Roberto Navone <ron@object65.it>
 
This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; version 2
of the License.
 
This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.
 
You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
 
*/

#include <nds.h>
#include <stdio.h>

#include "cwds.h"

#include "common.h"
#include "recognize.h"

#include "cwsamples_dat.h"

/* Externals for cwmain.c */

int profile_open_write(char *filename);
int profile_open_read(unsigned char *smpData, unsigned long smpDataLen);
void profile_close();
int profile_more_samples();
void set_logfilename(char *filename);

/* Externals for recognize.c */

void recognize_init(void);
void samples_write(void);
void sample_read(void);
void update_enabled_samples(void);

/* From cellwidget.c */

#define ALTERNATES 5

#define CELL_SHOW_INK   0x01
#define CELL_DIRTY      0x02
#define CELL_VERIFIED   0x04
#define CELL_SHIFTED    0x08

struct Cell
{
  Sample sample, *alts[ALTERNATES];
  gunichar2 ch;
  int alt_used[ALTERNATES];
  char flags, alt_ratings[ALTERNATES];
};

static struct Cell wrkCell;
static bool drawing = FALSE;

/* From recognize.c */

extern int msec;

/* Private vars & definitions */

#define MODE_None       0
#define MODE_Training   1
#define MODE_Recognize  2

static short mode;
static int cellw, cellh;
static int cellch;

static void draw(int x, int y);
static void stop_drawing();
static void finish_cell();

/*
 * Library initialization
 */
void cwdsInit(unsigned char *smpData, unsigned long smpDataLen)
{
  /* Set default samples data, if not provided */
  
  if (smpData == 0)
  {
    smpData = (unsigned char *)cwsamples_dat;
    smpDataLen = cwsamples_dat_size;
  }
  
  /* CW initialization */
  
  mode = MODE_None;
  
  recognize_init();
  
  /* Set filename for logging */
  
  set_logfilename("fat:/cellwriter.log");
  
  /* Read profile info */
  
  profile_open_read(smpData, smpDataLen);
  
  while (profile_more_samples())
    sample_read();
    
  profile_close();
  
  /* Enable samples */
  
  update_enabled_samples();
}

/*
 * Enter in recognize mode
 *
 * w width of cell area
 * h height of cell area
 */
void cwdsBeginRecognizeMode(int w, int h)
{
  mode = MODE_Recognize;
  
  cellch = 0;
  cellw = w;
  cellh = h;
}

/*
 * Notify a pen down event in cella area
 *
 * x, y event coordinates
 */
void cwdsNotifyPenDownEvent(int x, int y)
{
  /* If we are drawing and we get a button press event it is possible
     that we never received a button release event for some reason.
     This is a fix for Zaurus drawing connected lines. */
     
  if (drawing)
    stop_drawing();
    
  /* Draw with left click */
  
  draw(x, y);
}

/*
 * Notify a pen move event in cell area 
 *
 * x, y event coordinates
 */
void cwdsNotifyPenMoveEvent(int x, int y)
{
  /* Record and draw new segment */
  
  if (drawing) 
  {
    draw(x, y);
    
    //int stroke = input->len - 1;
    //int seg = input->strokes[input->len - 1]->len - 2;
    
    //render_segment((int)input->strokes[stroke]->points[seg].x, (int)input->strokes[stroke]->points[seg].y,
    //               (int)input->strokes[stroke]->points[seg + 1].x, (int)input->strokes[stroke]->points[seg + 1].y,
    //               m_cellX, m_cellY, m_cellW, m_cellH, m_color);
                   
    //render_segment(input, current_cell, input->len - 1,
    //               input->strokes[input->len - 1]->len - 2, NULL);
  }
}

/*
 * Notify a pen up event in cell area
 */
void cwdsNotifyPenUpEvent()
{
  stop_drawing();
}

/*
 * Terminate recognize mode and perform recognition
 *
 * returns the recognized symbol, -1 in no recognition done
 */
int cwdsEndRecognizeMode()
{
  finish_cell();
  
  mode = MODE_None;

  return (wrkCell.ch > 0 ? wrkCell.ch : -1);
}

/*
 * Abandon the recognize mode 
 */
void cwdsCancelRecognizeMode()
{
  mode = MODE_None;
}

static void cwdsEnableDisableSymbol(int ch, unsigned char value)
{
  Sample *sample;

  sampleiter_reset();
  
  while ((sample = sampleiter_next())) 
  {
    if (sample->ch == ch)
      sample->enabled = value;
  }
}

static void cwdsEnableDisableSymbolRange(int chFrom, int chTo, unsigned char value)
{
  Sample *sample;

  sampleiter_reset();
  
  while ((sample = sampleiter_next())) 
  {
    if (sample->ch >= chFrom  &&  sample->ch <= chTo)
      sample->enabled = value;
  }
}

/*
 * Enable the requested symbol
 */
void cwdsEnableSymbol(int ch)
{
  cwdsEnableDisableSymbol(ch, TRUE);
}

/*
 * Enable symbols in requeste range
 */
void cwdsEnableSymbolRange(int chFrom, int chTo)
{
  cwdsEnableDisableSymbolRange(chFrom, chTo, TRUE);
}

/*
 * Disable the requested symbol
 */
void cwdsDisableSymbol(int ch)
{
  cwdsEnableDisableSymbol(ch, FALSE);
}

/*
 * Disable symbols in requested range
 */
void cwdsDisableSymbolRange(int chFrom, int chTo)
{
  cwdsEnableDisableSymbolRange(chFrom, chTo, FALSE);
}

/*
 * Returns the max number of recognition alternatives
 */
int cwdsGetMaxRecognitionAlts()
{
  return (ALTERNATES);
}

/*
 * Return the time spent for the last recognition, in msec
 */
int cwdsGetLastRecognitionTime()
{
  return (msec);
}

/*
 * Get the number of possible alternatives in the last recognition phase
 */
int cwdsGetLastRecognitionNumAlts()
{
  int i, n = 0;
  
  for (i = 0; i < ALTERNATES &&  wrkCell.alts[i]; i++) 
  {
    if (!sample_valid(wrkCell.alts[i], wrkCell.alt_used[i]))
      continue;
      
    n++;
  }
  
  return (n);
}

/*
 * Get informations about the requested alternative for the last recognition phase
 *
 * nAlt the alternative 
 *
 * returns ch the alternative symbol
 * returns nSample the sample used for the recognition
 * returns pct the percentage for the recognition
 */
void cwdsGetLastRecognitionAltInfo(int nAlt, int *ch, int *nSample, int *pct)
{
  int i, pos;
  
  for (i = 0, pos = 0; i < ALTERNATES &&  wrkCell.alts[i]; i++) 
  {
    if (!sample_valid(wrkCell.alts[i], wrkCell.alt_used[i]))
      continue;
      
    if (pos++ == nAlt)
    {
      *ch = wrkCell.alts[i]->ch;
      *pct = wrkCell.alt_ratings[i];

      /* Search the sample used for recognition */
      
      sampleiter_reset();
  
      *nSample = 0;
  
      Sample *sample;
  
      while ((sample = sampleiter_next()) != NULL)
      {
        if (sample == wrkCell.alts[i])
          return;   
          
        if (sample->ch == *ch)
          (*nSample)++;
      }

      return;
    }    
  }
}

/*
 * Enter in training mode
 *
 * ch the symbol to train
 * w width of cell area
 * h height of cell area
 */
void cwdsBeginTrainingMode(int ch, int w, int h)
{
  mode = MODE_Training;
  
  cellch = ch;
  cellw = w;
  cellh = h;
}

/*
 * Terminate training mode and store symbol definition
 */
void cwdsEndTrainingMode()
{
  finish_cell();
  
  mode = MODE_None;
}

/*
 * Abandon the training mode 
 */
void cwdsCancelTrainingMode()
{
  mode = MODE_None;
}

/* 
 * Sava samples
 */
void cwdsSaveSamples(char *fileName)
{
  if (profile_open_write(fileName) != TRUE)
    return;
    
  samples_write();
  
  profile_close();
}

/*
 * Returns the number of samples defined for the requested symbol
 */
int cwdsGetNumSamples(int ch)
{
  sampleiter_reset();
  
  int n = 0;
  
  Sample *sample;
  
  while ((sample = sampleiter_next()) != NULL)
    if (sample->ch == ch)
      n++;
      
  return (n);
}

/*
 * Reset all the samples defined for the requested symbol
 */
void cwdsResetSamples(int ch)
{
  untrain_char(ch);
}

/*
 * Get the number of graphics commands for the requested sample of the requested symbol
 */
int cwdsGetSampleNumGraphicsCommands(int ch, int nSample)
{
  /* Find sample to draw */
  
  sampleiter_reset();
  
  int n = 0;
  
  Sample *sample;
  
  while ((sample = sampleiter_next()) != NULL)
    if (sample->ch == ch)
      if (nSample == n++)
        break;
      
  if (sample == NULL)
    return (0);
    
  /* Count sample's drawing elements */
  
  int ncmd = 0;
  
  int i, j;
  
  for (i = 0; i < ((Sample *)sample)->len; i++)
    if (((Sample *)sample)->strokes[i]->len <= 1 ||  ((Sample *)sample)->strokes[i]->spread < DOT_SPREAD)
      ncmd++;
    else
    {
      for (j = 0; j < ((Sample *)sample)->strokes[i]->len - 1; j++)
        ncmd++;
    }
    
  return (ncmd);
}

/*
 * Returns the requeste graphic command for the requested sample of the requested symbol
 *
 * w the cell's drawing area width
 * h the cell's drawing area height
 *
 * returns graphic command coordinates within the cell's area:
 * if x2, y2 are -1 the command is a point (x1, y1) else the command is a line (x1, y1 to x2, y2)
 */
void cwdsGetSampleGraphicCommand(int ch, int nSample, int nCommand, int w, int h, int *_x1, int *_y1, int *_x2, int *_y2)
{
  /* Find sample to draw */
  
  sampleiter_reset();
  
  int n = 0;
  
  Sample *sample;
  
  while ((sample = sampleiter_next()) != NULL)
    if (sample->ch == ch)
      if (nSample == n++)
        break;
      
  if (sample == NULL)
    return;
    
  /* Return the requested command */
  
  int ncmd = 0;
  
  int i, j, x1, y1, x2, y2;
  
  for (i = 0; i < ((Sample *)sample)->len; i++)
    if (((Sample *)sample)->strokes[i]->len <= 1 ||  ((Sample *)sample)->strokes[i]->spread < DOT_SPREAD)
    {
      if (ncmd++ == nCommand)
      {
        x1 = ((Sample *)sample)->strokes[i]->points[0].x;
        y1 = ((Sample *)sample)->strokes[i]->points[0].y;
        
        *_x1 = w / 2 + x1 * w / SCALE;
        *_y1 = h / 2 + y1 * h / SCALE;
        
        *_x2 = -1;
        *_y2 = -1;

        return;
      }
    }
    else
    {
      for (j = 0; j < ((Sample *)sample)->strokes[i]->len - 1; j++)
      {
        if (ncmd++ == nCommand)
        {
          x1 = ((Sample *)sample)->strokes[i]->points[j].x;
          x2 = ((Sample *)sample)->strokes[i]->points[j + 1].x;
          y1 = ((Sample *)sample)->strokes[i]->points[j].y;
          y2 = ((Sample *)sample)->strokes[i]->points[j + 1].y;

          *_x1 = w / 2 + x1 * w / SCALE;
          *_y1 = h / 2 + y1 * h / SCALE;
        
          *_x2 = w / 2 + x2 * w / SCALE;
          *_y2 = h / 2 + y2 * h / SCALE;
        
          return;
        }
      }
    }   
}

/*
 * Draw a pixel
 */
static void draw(int x, int y)
{
  /* New character */
  
  if (!input || !input->len) 
  {
    clear_sample(&(wrkCell.sample));
    
    wrkCell.alts[0] = NULL;
    
    input = &(wrkCell.sample);
    
    wrkCell.sample.ch = cellch;
  }

  /* Allocate a new stroke if we aren't already drawing */
  
  Stroke *s = input->strokes[input->len - 1];
  
  if (!drawing) 
  {
    if (input->len >= STROKES_MAX)
      return;
      
    s = input->strokes[input->len++]= stroke_new(0);
    
    drawing = true;
    
    //if (input->len == 1)
    //  render_cell(current_cell);
  }
  
  /* Normalize the input */
  
  x = (x - cellw / 2) * SCALE / cellw;
  y = (y - cellh / 2) * SCALE / cellh;
  
  draw_stroke(&input->strokes[input->len - 1], x, y);
}

/*
 * Ends the current stroke and applies various processing functions
 */
static void stop_drawing()
{
  Stroke *stroke;

  if (!drawing)
    return;
    
  drawing = false;
  
  if (!input || input->len >= STROKES_MAX)
    return;
    
  stroke = input->strokes[input->len - 1];
  
  smooth_stroke(stroke);
  
  simplify_stroke(stroke);
  
  process_stroke(stroke);
  
  //render_cell(current_cell);
  
  //render_sample(input, current_cell);
}

/* 
 * End cell drawing
 */
static void finish_cell()
{
  if (!input || input->len < 1)
    return;
    
  //wrkCell.flags |= CELL_DIRTY;

  if (mode == MODE_Training)
    train_sample(&(wrkCell.sample), TRUE);
  else if (mode == MODE_Recognize)
  {
    recognize_sample(input, wrkCell.alts, ALTERNATES);
    
    wrkCell.ch = input->ch;
    
    int i;
    
    for (i=0; i<ALTERNATES && wrkCell.alts[i]; i++) 
    {
      wrkCell.alt_ratings[i] = wrkCell.alts[i]->rating;
      wrkCell.alt_used[i] = wrkCell.alts[i]->used;
    }
  }

  input = NULL;
  drawing = false;
}
